Attribute VB_Name = "modCompiler"
Option Explicit

Private Declare Function ShellExecute Lib "shell32.dll" Alias "ShellExecuteA" (ByVal hWnd As Long, ByVal lpOperation As String, ByVal lpFile As String, ByVal lpParameters As String, ByVal lpDirectory As String, ByVal nShowCmd As Long) As Long
Private Const SW_SHOWNORMAL = 1

Type STRING_ITEM
    Text As String
    Offset As Long
End Type

Public IDataSection() As Byte
Public UDataSection As Long
Public CodeSection() As Byte
Public ImportSection() As Byte
Public ResourceSection() As Byte
Public Strings() As STRING_ITEM

Sub Compile()
    Dim sOutput As String, i As Integer
    With frmMain
        bError = False
        .SaveAll
        If .ActiveForm.Filename = "" Then
            Message "File is not saved."
            Exit Sub
        End If
        .MousePointer = 11
        .Progress 1, 6, "Initializing compiler..."
        InitCompiler
        sOutput = Left(.ActiveForm.Filename, InStrRev(.ActiveForm.Filename, ".")) & "exe"
        KillIfExists sOutput
        ChDir GetFilePath(.ActiveForm.Filename)
        If bError Then Exit Sub
        .Progress 2, 6, "Starting the compilation process..."
        sCode = .ActiveForm.cs.Text
        SkipBlank
        Application
        If bAppLib Then
            .Progress 3, 6, "Parsing library..."
            GlobalData
            If bError Then Exit Sub
            sOutput = Left(.ActiveForm.Filename, InStrRev(.ActiveForm.Filename, ".")) & "vl"
            ExportData sOutput
            Message "Library successfully compiled." & vbCrLf & FileLen(sOutput) & " bytes written."
        Else
            AddAlias "GetModuleHandle", "GetModuleHandleA", "KERNEL32.DLL", 1, False
            AddAlias "Copy", "lstrcpyA", "KERNEL32.DLL", 2, False
            AddAlias "CopyN", "lstrcpynA", "KERNEL32.DLL", 2, False
            AddAlias "Compare", "lstrcmpA", "KERNEL32.DLL", 2, False
            AddAlias "CompareI", "lstrcmpiA", "KERNEL32.DLL", 2, False
            AddAlias "Length", "lstrlenA", "KERNEL32.DLL", 1, False
            AddAlias "Append", "lstrcatA", "KERNEL32.DLL", 2, False
            AddAlias "Format", "wsprintfA", "USER32.DLL", -1, True
            AddAlias "MessageBox", "MessageBoxA", "USER32.DLL", 4, False
            AddAlias "ExitProcess", "ExitProcess", "KERNEL32.DLL", 1, False
            AddGlobalUData "Instance", 4, "dword"
            AddCodeWord &H6A
            Invoke "GetModuleHandle", 1
            AddCodeByte &HA3
            CodeFixup "Instance"
            Jump "$start"
            If bError Then Exit Sub
            .Progress 3, 6, "Parsing data section..."
            GlobalData
            Begin
            AddSymbol "$start", 0, stCode, stFunction, ""
            For i = 1 To UBound(LibBegin)
                StdCall LibBegin(i), 0
            Next i
            If bError Then Exit Sub
            .Progress 4, 6, "Parsing code section..."
            CodeBlock
            EndApp
            AddSymbol "$end", 0, stCode, stFunction, ""
            For i = 1 To UBound(LibEnd)
                StdCall LibEnd(i), 0
            Next i
            AddCodeWord &H6A
            Invoke "ExitProcess", 1
            If bError Then Exit Sub
            .Progress 5, 6, "Making executable file..."
            Link sOutput
            If bError Then Exit Sub
            Message "Successfully compiled." & vbCrLf & FileLen(sOutput) & " bytes written."
        End If
    End With
End Sub

Sub Run()
    ChDir GetFilePath(frmMain.ActiveForm.Filename)
    ShellExecute frmMain.hWnd, vbNullString, Left(frmMain.ActiveForm.Filename, InStrRev(frmMain.ActiveForm.Filename, ".")) & "exe", frmMain.ActiveForm.Commandline, GetFilePath(frmMain.ActiveForm.Filename), SW_SHOWNORMAL
End Sub

Sub InitCompiler()
    ReDim Preserve IDataSection(0) As Byte
    UDataSection = 0
    ReDim Preserve CodeSection(0) As Byte
    ReDim Preserve ImportSection(0) As Byte
    ReDim Preserve ResourceSection(0) As Byte
    ReDim Preserve Strings(0) As STRING_ITEM
    InitSymbols
    InitLinker
    InitFixups
    InitParser
    InitSymbolTypes
    InitResources
    sUsedLibs = ""
    bError = False
End Sub

Sub Message(sMessage As String)
    If bError Then Exit Sub
    frmMain.Progress 6, 6, ""
    frmMain.MousePointer = 0
    MsgBox sMessage
End Sub

Sub AddIDataByte(Value As Byte)
    ReDim Preserve IDataSection(UBound(IDataSection) + 1) As Byte
    IDataSection(UBound(IDataSection)) = Value
End Sub

Sub AddIDataWord(Value As Integer)
    AddIDataByte LoByte(Value)
    AddIDataByte HiByte(Value)
End Sub

Sub AddIDataDWord(Value As Long)
    AddIDataWord LoWord(Value)
    AddIDataWord HiWord(Value)
End Sub

Sub AddIDataString(Value As String)
    Dim i As Integer
    ReDim Preserve Strings(UBound(Strings) + 1) As STRING_ITEM
    With Strings(UBound(Strings))
        .Text = Value
        .Offset = UBound(IDataSection)
    End With
    For i = 1 To Len(Value)
        AddIDataByte CByte(Asc(mId(Value, i, 1)))
    Next i
    AddIDataByte 0
End Sub

Function TextPos(sText As String) As Long
    Dim i As Long
    TextPos = 0
    For i = 1 To UBound(Strings)
        If sText = Strings(i).Text Then
            TextPos = Strings(i).Offset + 1
            Exit Function
        End If
    Next i
End Function

Sub AddUDataByte()
    UDataSection = UDataSection + 1
End Sub

Sub AddUDataWord()
    UDataSection = UDataSection + 2
End Sub

Sub AddUDataDWord()
    UDataSection = UDataSection + 4
End Sub

Sub AddUDataBuffer(lSize As Long)
    UDataSection = UDataSection + lSize
End Sub

Sub AddCodeByte(Value As Byte)
    ReDim Preserve CodeSection(UBound(CodeSection) + 1) As Byte
    CodeSection(UBound(CodeSection)) = Value
End Sub

Sub AddCodeWord(Value As Integer)
    AddCodeByte LoByte(Value)
    AddCodeByte HiByte(Value)
End Sub

Sub AddCodeDWord(Value As Long)
    AddCodeWord LoWord(Value)
    AddCodeWord HiWord(Value)
End Sub

Sub AddImportByte(Value As Byte)
    ReDim Preserve ImportSection(UBound(ImportSection) + 1) As Byte
    ImportSection(UBound(ImportSection)) = Value
End Sub

Sub AddImportWord(Value As Integer)
    AddImportByte LoByte(Value)
    AddImportByte HiByte(Value)
End Sub

Sub AddImportDWord(Value As Long)
    AddImportWord LoWord(Value)
    AddImportWord HiWord(Value)
End Sub

Sub AddImportString(Value As String)
    Dim i As Integer
    For i = 1 To Len(Value)
        AddImportByte CByte(Asc(mId(Value, i, 1)))
    Next i
    AddImportByte 0
End Sub

Sub AddResourceByte(Value As Byte)
    ReDim Preserve ResourceSection(UBound(ResourceSection) + 1) As Byte
    ResourceSection(UBound(ResourceSection)) = Value
End Sub

Sub AddResourceWord(Value As Integer)
    AddResourceByte LoByte(Value)
    AddResourceByte HiByte(Value)
End Sub

Sub AddResourceDWord(Value As Long)
    AddResourceWord LoWord(Value)
    AddResourceWord HiWord(Value)
End Sub

Sub AddResourceString(Value As String)
    Dim i As Integer
    For i = 1 To Len(Value)
        AddResourceWord CInt(Asc(mId(Value, i, 1)))
    Next i
    AddResourceWord 0
End Sub

Sub ResourceAlign()
    Dim i As Long
    If UBound(ResourceSection) = Int(UBound(ResourceSection) / 4) * 4 Then Exit Sub
    For i = UBound(ResourceSection) To Int((UBound(ResourceSection) / 4) + 1) * 4 - 1
        AddResourceByte 0
    Next i
End Sub
