Attribute VB_Name = "modFixups"
Option Explicit

Private Type FIXUP_ITEM
    Symbol As String
    Section As SectionType
    Offset As Long
    Relative As Long
    ArgCount As Integer
    LineNumber As Long
    Filename As String
    ExtraAdd As Long
    ExtraSub As String
End Type

Dim Fixups() As FIXUP_ITEM

Sub InitFixups()
    ReDim Preserve Fixups(0) As FIXUP_ITEM
End Sub

Sub AddFixup(Symbol As String, Section As SectionType, Offset As Long, Relative As Long, Optional ArgCount As Integer = -1, Optional lAdd As Long = 0, Optional sSub As String = "")
    ReDim Preserve Fixups(UBound(Fixups) + 1) As FIXUP_ITEM
    If SymbolExists(sCurrentProc & "." & Symbol) Then Symbol = sCurrentProc & "." & Symbol
    With Fixups(UBound(Fixups))
        .Symbol = Symbol
        .Section = Section
        .Offset = Offset
        .Relative = Relative
        .ArgCount = ArgCount
        .LineNumber = lCurrentLine
        .Filename = sCurrentFile
        .ExtraAdd = lAdd
        .ExtraSub = sSub
    End With
End Sub

Sub ResourceFixup(Symbol As String, lAdd As Long, sSub As String)
    AddFixup Symbol, stResource, UBound(ResourceSection) + 1, 0, -1, lAdd, sSub
    AddResourceDWord 0
End Sub

Sub ResourceFixupWord(Symbol As String)
    AddFixup Symbol, stResource, UBound(ResourceSection) + 1, 0, , , "WORD"
    AddResourceWord 0
End Sub

Sub CodeFixup(Symbol As String, Optional Relative As Long = 0, Optional ArgCount As Integer = -1)
    AddFixup Symbol, stCode, UBound(CodeSection) + 1, Relative, ArgCount
    AddCodeDWord 0
End Sub

Sub DoFixups()
    Dim i As Long, tmp As SYMBOL_ITEM, lOffset As Long, tmp2 As SYMBOL_ITEM
    For i = 1 To UBound(Fixups)
        With Fixups(i)
            lCurrentLine = .LineNumber
            sCurrentFile = .Filename
            tmp = GetSymbol(.Symbol)
            If tmp.sType <> stFunction And .Relative Then errCallVariable tmp.Name
            lOffset = GetVirtualOffset(tmp, .Relative)
            lOffset = lOffset + .ExtraAdd
            If .ExtraSub = "RVA" Then
                lOffset = lOffset - &H400000
            ElseIf .ExtraSub <> "" And .ExtraSub <> "WORD" Then
                tmp2 = GetSymbol(.ExtraSub)
                lOffset = lOffset - GetVirtualOffset(tmp2, 0)
            End If
            If .ArgCount <> tmp.ArgCount And tmp.ArgCount >= 0 And .ArgCount >= 0 Then
                errInvalidParamCount
            End If
            Select Case .Section
            Case stIData
                IDataSection(.Offset) = LoByte(LoWord(lOffset))
                IDataSection(.Offset + 1) = HiByte(LoWord(lOffset))
                IDataSection(.Offset + 2) = LoByte(HiWord(lOffset))
                IDataSection(.Offset + 3) = HiByte(HiWord(lOffset))
            Case stCode
                CodeSection(.Offset) = LoByte(LoWord(lOffset))
                CodeSection(.Offset + 1) = HiByte(LoWord(lOffset))
                CodeSection(.Offset + 2) = LoByte(HiWord(lOffset))
                CodeSection(.Offset + 3) = HiByte(HiWord(lOffset))
            Case stImport
                ImportSection(.Offset) = LoByte(LoWord(lOffset))
                ImportSection(.Offset + 1) = HiByte(LoWord(lOffset))
                ImportSection(.Offset + 2) = LoByte(HiWord(lOffset))
                ImportSection(.Offset + 3) = HiByte(HiWord(lOffset))
            Case stResource
                If .ExtraSub = "WORD" Then lOffset = tmp.Offset
                ResourceSection(.Offset) = LoByte(LoWord(lOffset))
                ResourceSection(.Offset + 1) = HiByte(LoWord(lOffset))
                If .ExtraSub <> "WORD" Then
                    ResourceSection(.Offset + 2) = LoByte(HiWord(lOffset))
                    ResourceSection(.Offset + 3) = HiByte(HiWord(lOffset))
                End If
            End Select
        End With
    Next i
End Sub

Private Function GetVirtualOffset(Symbol As SYMBOL_ITEM, Relative As Long)
    With Symbol
        Select Case .Section
        Case stIData
            GetVirtualOffset = &H401000 + .Offset
        Case stUData
            GetVirtualOffset = &H401000 + UBound(IDataSection) + .Offset
        Case stCode
            If Relative <> 0 Then Relative = Relative + &H401000 + VirtualSizeOfData
            GetVirtualOffset = &H401000 + VirtualSizeOfData + .Offset - Relative
        Case stImport
            GetVirtualOffset = &H401000 + VirtualSizeOfData + VirtualSizeOfCode + .Offset
        Case stResource
            GetVirtualOffset = &H401000 + VirtualSizeOfData + VirtualSizeOfCode + VirtualSizeOfImport + .Offset
        End Select
    End With
End Function
