Attribute VB_Name = "modLinker"
Option Explicit

Private Type IMAGE_FILE_HEADER
    Machine As Integer
    NumberOfSections As Integer
    TimeDateStamp As Long
    PointerToSymbolTable As Long
    NumberOfSymbols As Long
    SizeOfOptionalHeader As Integer
    Characteristics As Integer
End Type

Private Type IMAGE_DATA_DIRECTORY
    VirtualAddress As Long
    Size As Long
End Type

Private Type IMAGE_OPTIONAL_HEADER
    Magic As Integer
    MajorLinkerVersion As Byte
    MinorLinkerVersion As Byte
    SizeOfCode As Long
    SizeOfInitializedData As Long
    SizeOfUninitializedData As Long
    AddressOfEntryPoint As Long
    BaseOfCode As Long
    BaseOfData As Long
    ImageBase As Long
    SectionAlignment As Long
    FileAlignment As Long
    MajorOperatingSystemVersion As Integer
    MinorOperatingSystemVersion As Integer
    MajorImageVersion As Integer
    MinorImageVersion As Integer
    MajorSubsystemVersion As Integer
    MinorSubsystemVersion As Integer
    Win32VersionValue As Long
    SizeOfImage As Long
    SizeOfHeaders As Long
    CheckSum As Long
    Subsystem As Integer
    DllCharacteristics As Integer
    SizeOfStackReserve As Long
    SizeOfStackCommit As Long
    SizeOfHeapReserve As Long
    SizeOfHeapCommit As Long
    LoaderFlags As Long
    NumberOfRvaAndSizes As Long
    DataDirectory(0 To 15) As IMAGE_DATA_DIRECTORY
End Type

Private Type IMAGE_SECTION_HEADER
    ImageName(0 To 7) As Byte
    VirtualSize As Long
    VirtualAddress As Long
    SizeOfRawData As Long
    PointerToRawData As Long
    PointerToRelocations As Long
    PointerToLinenumbers As Long
    NumberOfRelocations As Integer
    NumberOfLinenumbers As Integer
    Characteristics As Long
End Type

Private Type IMAGE_HEADER
    Signature As Long
    FileHeader As IMAGE_FILE_HEADER
    OptionalHeader As IMAGE_OPTIONAL_HEADER
End Type

Private Type IMPORT_ITEM
    Function As String
    Library As String
End Type

Private Type EXPORT_ITEM
    Function As String
    Name As String
End Type

Private Type IMPORT_DIRECTORY
    Name As String
    Size As Long
    Functions As Long
    SizeOfFunctions As Long
End Type

Public Enum ApplicationType
    atConsole = 3
    atGUI = 2
End Enum

Const FILE_SHARE_READ = &H1
Const FILE_SHARE_WRITE = &H2
Const CREATE_NEW = 1
Const GENERIC_WRITE = &H40000000

Declare Function CreateFile Lib "kernel32" Alias "CreateFileA" (ByVal lpFileName As String, ByVal dwDesiredAccess As Long, ByVal dwShareMode As Long, lpSecurityAttributes As Long, ByVal dwCreationDisposition As Long, ByVal dwFlagsAndAttributes As Long, ByVal hTemplateFile As Long) As Long
Declare Function WriteFile Lib "kernel32" (ByVal hFile As Long, lpBuffer As Any, ByVal nNumberOfBytesToWrite As Long, lpNumberOfBytesWritten As Long, lpOverlapped As Long) As Long
Declare Function CloseHandle Lib "kernel32" (ByVal hObject As Long) As Long

Private hFile As Long, FilePosition As Long
Private PEHeader As IMAGE_HEADER
Private Sections(0 To 3) As IMAGE_SECTION_HEADER
Private Imports() As IMPORT_ITEM
Private Exports() As EXPORT_ITEM
Public AppType As ApplicationType

Sub InitLinker()
    ReDim Preserve Imports(0) As IMPORT_ITEM
    ReDim Preserve Exports(0) As EXPORT_ITEM
    FilePosition = 0
    AppType = atGUI
End Sub

Sub Link(sFile As String)
    Dim BW As Long
    
' Finalize
    ImportExternalFunctions
    DoFixups
    If bError Then Exit Sub
    
' Create file
    hFile = CreateFile(sFile, GENERIC_WRITE, FILE_SHARE_READ Or FILE_SHARE_WRITE, ByVal 0&, CREATE_NEW, 0, 0)
    
' DOS MZ header
    OutputDWord &H805A4D:   OutputDWord &H1&:       OutputDWord &H100004:   OutputDWord &HFFFF&
    OutputDWord &H140&:     OutputDWord &H0&:       OutputDWord &H40&:      OutputDWord &H0&
    OutputDWord &H0&:       OutputDWord &H0&:       OutputDWord &H0&:       OutputDWord &H0&
    OutputDWord &H0&:       OutputDWord &H0&:       OutputDWord &H0&:       OutputDWord &H80&
' DOS stub
    OutputDWord &HEBA1F0E:  OutputDWord &HCD09B400: OutputDWord &H4C01B821: OutputDWord &H685421CD
    OutputDWord &H70207369: OutputDWord &H72676F72: OutputDWord &H63206D61: OutputDWord &H6F6E6E61
    OutputDWord &H65622074: OutputDWord &H6E757220: OutputDWord &H206E6920: OutputDWord &H20534F44
    OutputDWord &H65646F6D: OutputDWord &H240A0D2E: OutputDWord &H0&:       OutputDWord &H0&
    
' PE header
    With PEHeader
        .Signature = &H4550&
    End With
    With PEHeader.FileHeader
        .Machine = &H14C
        .NumberOfSections = NumberOfSections
        .SizeOfOptionalHeader = &HE0
        .Characteristics = &H818F
        .TimeDateStamp = 0
    End With
    With PEHeader.OptionalHeader
        .Magic = 267
        .MajorLinkerVersion = 1
        .MinorLinkerVersion = 0
        .Win32VersionValue = 2
        .AddressOfEntryPoint = &H1000 + VirtualSizeOfData
        .BaseOfCode = &H1000 + VirtualSizeOfData
        .BaseOfData = &H1000
        .ImageBase = &H400000
        .SectionAlignment = &H1000
        .FileAlignment = &H200
        .MajorOperatingSystemVersion = 1
        .MajorSubsystemVersion = 4
        .Win32VersionValue = 2
        .SizeOfImage = &H1000 + VirtualSizeOfData + VirtualSizeOfCode + VirtualSizeOfImport + VirtualSizeOfResource
        If PEHeader.FileHeader.NumberOfSections = 4 Then
            .SizeOfHeaders = &H400
        Else
            .SizeOfHeaders = &H200
        End If
        .SizeOfCode = UBound(CodeSection)
        .SizeOfInitializedData = UBound(IDataSection)
        .SizeOfUninitializedData = UDataSection
        .Subsystem = AppType
        .SizeOfStackCommit = &H10000
        .SizeOfStackReserve = &H10000
        .SizeOfHeapReserve = &H10000
        .NumberOfRvaAndSizes = 16
        If UBound(ImportSection) > 0 Then
            .DataDirectory(1).VirtualAddress = &H1000 + VirtualSizeOfData + VirtualSizeOfCode
            .DataDirectory(1).Size = UBound(ImportSection)
        Else
            .DataDirectory(1).VirtualAddress = 0
            .DataDirectory(1).Size = 0
        End If
        If UBound(ResourceSection) > 0 Then
            .DataDirectory(2).VirtualAddress = &H1000 + VirtualSizeOfData + VirtualSizeOfCode + VirtualSizeOfImport
            .DataDirectory(2).Size = UBound(ResourceSection)
        Else
            .DataDirectory(2).VirtualAddress = 0
            .DataDirectory(2).Size = 0
        End If
    End With
    OutputHeader PEHeader
    
' Section table
    ' .data
    If UBound(IDataSection) > 0 Or UDataSection > 0 Then
    With Sections(0)
        .ImageName(0) = 46
        .ImageName(1) = 100
        .ImageName(2) = 97
        .ImageName(3) = 116
        .ImageName(4) = 97
        .VirtualAddress = &H1000
        .VirtualSize = UBound(IDataSection) + UDataSection
        .SizeOfRawData = PhysicalSizeOfData
        .PointerToRawData = PEHeader.OptionalHeader.SizeOfHeaders
        .Characteristics = &HC0000040
    End With
    OutputSection Sections(0)
    End If
    ' .code
    If UBound(CodeSection) > 0 Then
    With Sections(1)
        .ImageName(0) = 46
        .ImageName(1) = 99
        .ImageName(2) = 111
        .ImageName(3) = 100
        .ImageName(4) = 101
        .VirtualAddress = &H1000 + VirtualSizeOfData
        .VirtualSize = UBound(CodeSection)
        .SizeOfRawData = PhysicalSizeOfCode
        .PointerToRawData = PEHeader.OptionalHeader.SizeOfHeaders + PhysicalSizeOfData
        .Characteristics = &H60000020
    End With
    OutputSection Sections(1)
    End If
    ' .idata
    If UBound(ImportSection) > 0 Then
    With Sections(2)
        .ImageName(0) = 46
        .ImageName(1) = 105
        .ImageName(2) = 100
        .ImageName(3) = 97
        .ImageName(4) = 116
        .ImageName(5) = 97
        .VirtualAddress = &H1000 + VirtualSizeOfData + VirtualSizeOfCode
        .VirtualSize = UBound(ImportSection)
        .SizeOfRawData = PhysicalSizeOfImport
        .PointerToRawData = PEHeader.OptionalHeader.SizeOfHeaders + PhysicalSizeOfData + PhysicalSizeOfCode
        .Characteristics = &HC0000040
    End With
    OutputSection Sections(2)
    End If
    ' .rsrc
    If UBound(ResourceSection) > 0 Then
    With Sections(3)
        .ImageName(0) = 46
        .ImageName(1) = 114
        .ImageName(2) = 115
        .ImageName(3) = 114
        .ImageName(4) = 99
        .VirtualAddress = &H1000 + VirtualSizeOfData + VirtualSizeOfCode + VirtualSizeOfImport
        .VirtualSize = UBound(ResourceSection)
        .SizeOfRawData = PhysicalSizeOfResource
        .PointerToRawData = PEHeader.OptionalHeader.SizeOfHeaders + PhysicalSizeOfData + PhysicalSizeOfCode + PhysicalSizeOfImport
        .Characteristics = &H40000040
    End With
    OutputSection Sections(3)
    End If
    Align
    
' Data section
    If UBound(IDataSection) > 0 Or UDataSection > 0 Then
    WriteFile hFile, IDataSection(1), UBound(IDataSection), BW, ByVal 0&
    FilePosition = FilePosition + UBound(IDataSection)
    Align
    End If
    
' Code section
    If UBound(CodeSection) > 0 Then
    WriteFile hFile, CodeSection(1), UBound(CodeSection), BW, ByVal 0&
    FilePosition = FilePosition + UBound(CodeSection)
    Align
    End If
    
' Import data section
    If UBound(ImportSection) > 0 Then
    WriteFile hFile, ImportSection(1), UBound(ImportSection), BW, ByVal 0&
    FilePosition = FilePosition + UBound(ImportSection)
    Align
    End If
    
' Resource section
    If UBound(ResourceSection) > 0 Then
    WriteFile hFile, ResourceSection(1), UBound(ResourceSection), BW, ByVal 0&
    FilePosition = FilePosition + UBound(ResourceSection)
    Align
    End If

' Close file
    CloseHandle hFile
End Sub

Private Sub OutputByte(Value As Byte)
    Dim BW As Long
    WriteFile hFile, Value, 1, BW, ByVal 0&
    FilePosition = FilePosition + 1
End Sub

Private Sub OutputWord(Value As Integer)
    Dim BW As Long
    WriteFile hFile, Value, 2, BW, ByVal 0&
    FilePosition = FilePosition + 2
End Sub

Private Sub OutputDWord(Value As Long)
    Dim BW As Long
    WriteFile hFile, Value, 4, BW, ByVal 0&
    FilePosition = FilePosition + 4
End Sub

Private Sub OutputString(Value As String)
    Dim BW As Long
    WriteFile hFile, Value, Len(Value), BW, ByVal 0&
    OutputByte 0
    FilePosition = FilePosition + Len(Value) + 1
End Sub

Private Sub OutputHeader(Value As IMAGE_HEADER)
    Dim BW As Long
    WriteFile hFile, Value, LenB(Value), BW, ByVal 0&
    FilePosition = FilePosition + LenB(Value)
End Sub

Private Sub OutputSection(Value As IMAGE_SECTION_HEADER)
    Dim BW As Long
    WriteFile hFile, Value, LenB(Value), BW, ByVal 0&
    FilePosition = FilePosition + LenB(Value)
End Sub

Private Sub Align()
    Dim i As Long, ii As Long
    For i = 0 To 512 * 100000 Step 512
        If i > FilePosition Then
            For ii = 1 To i - FilePosition
                OutputByte 0
            Next ii
            Exit For
        End If
    Next i
End Sub

Private Function SectionPhysicalSize(Value() As Byte) As Long
    Dim i As Long
    If UBound(Value) = 0 Then
        SectionPhysicalSize = 0
    Else
        For i = 512 To 512 * 100000 Step 512
            If i > UBound(Value) Then
                SectionPhysicalSize = i
                Exit For
            End If
        Next i
    End If
End Function

Private Function SectionVirtualSize(Value() As Byte, Optional Extra As Long = 0) As Long
    Dim i As Long
    If UBound(Value) = 0 Then
        SectionVirtualSize = 0
    Else
        For i = &H1000 To &H1000& * &HFFFF& Step &H1000
            If i > (UBound(Value) + Extra) Then
                SectionVirtualSize = i
                Exit For
            End If
        Next i
    End If
End Function

Private Function SectionVirtualSizeEx(Size As Long) As Long
    Dim i As Long
    For i = &H1000 To &H1000& * &HFFFF& Step &H1000
        If i > Size Then
            SectionVirtualSizeEx = i
            Exit For
        End If
    Next i
End Function

Function PhysicalSizeOfData() As Long
    PhysicalSizeOfData = SectionPhysicalSize(IDataSection)
End Function

Function PhysicalSizeOfCode() As Long
    PhysicalSizeOfCode = SectionPhysicalSize(CodeSection)
End Function

Function PhysicalSizeOfImport() As Long
    PhysicalSizeOfImport = SectionPhysicalSize(ImportSection)
End Function

Function PhysicalSizeOfResource() As Long
    PhysicalSizeOfResource = SectionPhysicalSize(ResourceSection)
End Function

Function VirtualSizeOfData() As Long
    VirtualSizeOfData = SectionVirtualSize(IDataSection, UDataSection)
End Function

Function VirtualSizeOfCode() As Long
    VirtualSizeOfCode = SectionVirtualSize(CodeSection)
End Function

Function VirtualSizeOfImport() As Long
    VirtualSizeOfImport = SectionVirtualSize(ImportSection)
End Function

Function VirtualSizeOfResource() As Long
    VirtualSizeOfResource = SectionVirtualSize(ResourceSection)
End Function

Function NumberOfSections() As Long
    NumberOfSections = 0
    If (UBound(IDataSection) + UDataSection) > 0 Then
        NumberOfSections = NumberOfSections + 1
    End If
    If UBound(CodeSection) > 0 Then
        NumberOfSections = NumberOfSections + 1
    End If
    If UBound(ImportSection) > 0 Then
        NumberOfSections = NumberOfSections + 1
    End If
    If UBound(ResourceSection) > 0 Then
        NumberOfSections = NumberOfSections + 1
    End If
End Function

Sub Import(sFunction As String, sLibrary As String)
    Dim i As Long
    For i = 1 To UBound(Imports)
        If Imports(i).Function = sFunction And Imports(i).Library = sLibrary Then Exit Sub
    Next i
    ReDim Preserve Imports(UBound(Imports) + 1) As IMPORT_ITEM
    With Imports(UBound(Imports))
        .Function = sFunction
        .Library = sLibrary
    End With
End Sub

Sub Export(sFunction As String, sName As String)
    Dim i As Long
    For i = 1 To UBound(Exports)
        If Exports(i).Function = sFunction And Exports(i).Name = sName Then Exit Sub
    Next i
    ReDim Preserve Exports(UBound(Exports) + 1) As EXPORT_ITEM
    With Exports(UBound(Exports))
        .Function = sFunction
        .Name = sName
    End With
End Sub

Private Sub ImportExternalFunctions()
    Dim Directory() As IMPORT_DIRECTORY
    Dim SizeOfImportHeader As Long, i As Integer, ii As Integer
    Dim lPointerToName As Integer, bDirExists As Boolean
    ReDim Preserve Directory(0) As IMPORT_DIRECTORY
    
' Quit if none imports
    If UBound(Imports) = 0 Then Exit Sub
    
' Sorting the libraries and get the size of each of them
    For i = 1 To UBound(Imports)
        bDirExists = False
        For ii = 1 To UBound(Directory)
            If Imports(i).Library = Directory(ii).Name Then
                bDirExists = True
            End If
        Next ii
        If Not bDirExists Then
            ReDim Preserve Directory(UBound(Directory) + 1) As IMPORT_DIRECTORY
            With Directory(UBound(Directory))
                .Name = Imports(i).Library
                .Size = 0
                .SizeOfFunctions = 4
                .Functions = 0
            End With
        Else
            For ii = 1 To UBound(Directory)
                If Directory(ii).Name = Imports(i).Library Then
                    Directory(ii).Size = Directory(ii).Size + 7 + Len(Imports(i).Function)
                End If
            Next ii
        End If
    Next i
    
' Calculate size of library table
    SizeOfImportHeader = 20
    For i = 1 To UBound(Directory)
        SizeOfImportHeader = SizeOfImportHeader + 20 + Len(Directory(i).Name) + 1
    Next i
    
' Calculate size of function table
    For i = 1 To UBound(Directory)
        For ii = 1 To UBound(Imports)
            If Imports(ii).Library = Directory(i).Name Then
                Directory(i).SizeOfFunctions = Directory(i).SizeOfFunctions + Len(Imports(ii).Function) + 7
                Directory(i).Functions = Directory(i).Functions + 1
            End If
        Next ii
    Next i
    
' Output libraries
    lPointerToName = 0
    For i = 1 To UBound(Directory)
        AddImportDWord 0
        AddImportDWord 0
        AddImportDWord 0
        AddImportDWord &H1000 + VirtualSizeOfData + VirtualSizeOfCode + ((UBound(Directory) + 1) * 20) + lPointerToName
        AddImportDWord &H1000 + VirtualSizeOfData + VirtualSizeOfCode + SizeOfImportHeader
        lPointerToName = lPointerToName + Len(Directory(i).Name) + 1
        SizeOfImportHeader = SizeOfImportHeader + Directory(i).SizeOfFunctions
    Next i
    AddImportDWord 0
    AddImportDWord 0
    AddImportDWord 0
    AddImportDWord 0
    AddImportDWord 0
    For i = 1 To UBound(Directory)
        AddImportString Directory(i).Name
    Next i
    
' Output functions
    For i = 1 To UBound(Directory)
        lPointerToName = (Directory(i).Functions + 1) * 4
        For ii = 1 To UBound(Imports)
            If Imports(ii).Library = Directory(i).Name Then
                AddSymbol Imports(ii).Function & "@" & Imports(ii).Library, 0, stImport, stFunction, ""
                AddImportDWord &H1000 + VirtualSizeOfData + VirtualSizeOfCode + UBound(ImportSection) + lPointerToName
                lPointerToName = lPointerToName + Len(Imports(ii).Function) - 1
            End If
        Next ii
        AddImportDWord 0
        For ii = 1 To UBound(Imports)
            If Imports(ii).Library = Directory(i).Name Then
                AddImportWord 0
                AddImportString Imports(ii).Function
            End If
        Next ii
    Next i
End Sub
