Attribute VB_Name = "modParser"
Option Explicit

Public Type ALIAS_ITEM
    Alias As String
    Name As String
    Library As String
    ParamCount As Long
    PopParams As Boolean
End Type

Public Type CONSTANT_ITEM
    Name As String
    Value As Long
End Type

Public sCode As String, sCurrentFile As String, lCurrentLine As Long, lUniqueID As Long
Public sCurrentProc As String, lProcDataSize As Long
Public Alias() As ALIAS_ITEM, Constants() As CONSTANT_ITEM, LibBegin() As String, LibEnd() As String, lEnumCount As Long

Sub InitParser()
    lCurrentLine = 1: sCurrentFile = "": lUniqueID = 1
    ReDim Preserve Alias(0) As ALIAS_ITEM
    ReDim Preserve Constants(0) As CONSTANT_ITEM
    ReDim Preserve LibBegin(0) As String
    ReDim Preserve LibEnd(0) As String
    lEnumCount = 0
    AddConstant "TRUE", -1
    AddConstant "FALSE", 0
    AddConstant "NULL", 0
End Sub

Sub AddAlias(sAlias As String, sName As String, sLibrary As String, lParamCount As Long, bPopParams As Boolean)
    If AliasExists(sAlias) Or ConstantExists(sAlias) Or SymbolExists(sAlias) Then
        errIdentifierExists sAlias
        Exit Sub
    End If
    ReDim Preserve Alias(UBound(Alias) + 1) As ALIAS_ITEM
    With Alias(UBound(Alias))
        .Alias = sAlias
        .Name = sName
        .Library = sLibrary
        .ParamCount = lParamCount
        .PopParams = bPopParams
    End With
End Sub

Function AliasExists(sName As String) As Boolean
    Dim i As Long
    AliasExists = False
    For i = 1 To UBound(Alias)
        If Alias(i).Alias = sName Then
            AliasExists = True
            Exit Function
        End If
    Next i
End Function

Function GetAlias(sAlias As String) As ALIAS_ITEM
    Dim i As Long
    For i = 1 To UBound(Alias)
        If Alias(i).Alias = sAlias Then
            GetAlias = Alias(i)
            Exit Function
        End If
    Next i
End Function

Sub AddConstant(sName As String, lValue As Long)
    If AliasExists(sName) Or ConstantExists(sName) Or SymbolExists(sName) Then
        errIdentifierExists sName
        Exit Sub
    End If
    ReDim Preserve Constants(UBound(Constants) + 1) As CONSTANT_ITEM
    Constants(UBound(Constants)).Name = sName
    Constants(UBound(Constants)).Value = lValue
End Sub

Function ConstantExists(sName As String) As Boolean
    Dim i As Long
    ConstantExists = False
    For i = 1 To UBound(Constants)
        If Constants(i).Name = sName Then
            ConstantExists = True
            Exit Function
        End If
    Next i
End Function

Function GetConstant(sName As String) As Long
    Dim i As Long
    For i = 1 To UBound(Constants)
        If Constants(i).Name = sName Then
            GetConstant = Constants(i).Value
            Exit Function
        End If
    Next i
End Function

Sub AddEnum(sName As String)
    lEnumCount = lEnumCount + 1
    AddConstant sName, lEnumCount
End Sub

Function Locked(Optional count As Long = 1) As String
    Locked = Left(sCode, count)
End Function

Sub Skip(Optional count As Long = 1)
    sCode = mId(sCode, count + 1)
End Sub

Function IsBlank() As Boolean
    IsBlank = (Locked = " ") Or (Locked = vbTab)
End Function

Sub Blank()
    If Not IsBlank Then errUnexpected "blank"
    While IsBlank
        Skip
    Wend
End Sub

Sub SkipBlank()
    Dim bC1 As Boolean, bC2 As Boolean
    While True
        If IsBlank Or Locked = vbLf Then
            Skip
        ElseIf Locked = vbCr Then
            lCurrentLine = lCurrentLine + 1
            Skip
            bC1 = False
        ElseIf Locked(2) = "//" And Not bC2 Then
            bC1 = True
            Skip 2
        ElseIf Locked(2) = "/*" And Not bC1 Then
            bC2 = True
            Skip 2
        ElseIf Locked(2) = "*/" Then
            bC2 = False
            Skip 2
        ElseIf bC1 Or bC2 Then
            Skip
        Else
            Exit Sub
        End If
        If Len(sCode) = 0 Then Exit Sub
    Wend
End Sub

Function IsTerminator() As Boolean
    IsTerminator = IsSymbol(";")
End Function

Sub Terminator()
    If Not IsTerminator Then errUnexpected "terminator"
    While IsBlank
        Skip
    Wend
    Skip
    SkipBlank
End Sub

Function IsIdentifier(Optional ConstExists As Boolean = False) As Boolean
    Dim i As Long, sTemp As String
    i = 1: sTemp = ""
    While (UCase(mId(sCode, i, 1)) >= "A" And UCase(mId(sCode, i, 1)) <= "Z") Or _
    IIf(i > 1, (mId(sCode, i, 1) >= "0" And mId(sCode, i, 1) <= "9") Or mId(sCode, i, 1) = "." Or mId(sCode, i, 1) = "_", False)
        sTemp = sTemp & mId(sCode, i, 1)
        i = i + 1
    Wend
    IsIdentifier = (Len(sTemp) > 0) And IIf(ConstExists, ConstantExists(sTemp), True)
End Function

Function Identifier() As String
    If Not IsIdentifier Then errUnexpected "identifier"
    While (UCase(Locked) >= "A" And UCase(Locked) <= "Z") Or _
    (Locked >= "0" And Locked <= "9") Or Locked = "." Or Locked = "_"
        Identifier = Identifier & Locked
        Skip
    Wend
End Function

Function IsNumber() As Boolean
    IsNumber = IsIdentifier(True) Or IsSymbol("[") Or IsSymbol("'") Or (Locked >= "0" And Locked <= "9") Or _
        (Locked = "-" And mId(sCode, 2, 1) >= "0" And mId(sCode, 2, 1) <= "9") Or _
        IsSymbol("$")
End Function

Function Number() As Long
    Dim bIsNeg As Boolean
    If Not IsNumber Then errUnexpected "number"
    If IsIdentifier(True) Then
        Number = GetConstant(Identifier)
    ElseIf IsSymbol("[") Then
        Number = ConstantExpression
    ElseIf IsSymbol("'") Then
        Symbol "'"
        Number = Asc(Locked)
        Skip
        Symbol "'"
    ElseIf IsSymbol("$") Then
        Skip
        While (Locked >= "0" And Locked <= "9") Or LCase(Locked) >= "a" And LCase(Locked) <= "f"
            If LongToUnsigned(Number) >= (OFFSET_4 / 16) Then errValueOutOfRange
            Number = Multiply(Number, 16) + CharToDigit(Locked)
            Skip
        Wend
    Else
        bIsNeg = (Locked = "-")
        If bIsNeg Then Skip
        While Locked >= "0" And Locked <= "9"
            If LongToUnsigned(Number) >= (OFFSET_4 / 10) Then errValueOutOfRange
            Number = Multiply(Number, 10) + CharToDigit(Locked)
            Skip
        Wend
        If bIsNeg Then Number = 0 - Number
    End If
End Function

Function IsText() As Boolean
    IsText = (Locked = """")
End Function

Function Text() As String
    If Not IsText Then errUnexpected "string"
    Skip
    While Not Locked = """"
        If Locked = "\" Then
            Skip
            Select Case Locked
            Case "n": Text = Text & vbCrLf
            Case "t": Text = Text & vbTab
            Case """": Text = Text & """"
            Case "\": Text = Text & "\"
            Case Else: Text = Text & "\" & Locked
            End Select
            Skip
        ElseIf Locked(2) = vbCrLf Then
            errUnterminated
        Else
            Text = Text & Locked
            Skip
        End If
        If Len(sCode) = 0 Then Exit Function
    Wend
    Skip
End Function

Function IsSymbol(sSymbol As String) As Boolean
    Dim i As Long: i = 1
    While mId(sCode, i, 1) = " " Or mId(sCode, i, 1) = vbTab
        i = i + 1
    Wend
    IsSymbol = mId(sCode, i, Len(sSymbol)) = sSymbol
End Function

Sub Symbol(sSymbol As String, Optional SkipPostBlank As Boolean = True)
    While IsBlank
        Skip
    Wend
    If Not Locked(Len(sSymbol)) = sSymbol Then errUnexpected "symbol '" & sSymbol & "'"
    Skip Len(sSymbol)
    If Not SkipPostBlank Then Exit Sub
    While IsBlank
        Skip
    Wend
    If IsSymbol("_") Then
        Symbol "_"
        SkipBlank
    End If
End Sub
