Attribute VB_Name = "modSymbols"
Option Explicit

Public Enum SectionType
    stIData
    stUData
    stCode
    stImport
    stResource
End Enum

Public Enum SymbolType
    stFunction
    stGlobal
    stLocal
End Enum

Public Type SYMBOL_ITEM
    Name As String
    Offset As Long
    Index As Long
    Section As SectionType
    Size As Long
    sType As SymbolType
    TypeName As String
    Parameter As Boolean
    Pointer As Boolean
    UserDefined As String
    ArgCount As Integer
End Type

Public Symbols() As SYMBOL_ITEM

Sub InitSymbols()
    ReDim Preserve Symbols(0) As SYMBOL_ITEM
End Sub

Function IsResIdent(sIdent As String) As Boolean
    IsResIdent = (LCase(sIdent) = "application") Or (LCase(sIdent) = "resources") Or (LCase(sIdent) = "begin") Or (LCase(sIdent) = "end") Or (LCase(sIdent) = "end.") Or _
        (LCase(sIdent) = "alias") Or (LCase(sIdent) = "const") Or (LCase(sIdent) = "frame") Or (LCase(sIdent) = "if") Or (LCase(sIdent) = "elseif") Or _
        (LCase(sIdent) = "else") Or (LCase(sIdent) = "return") Or (LCase(sIdent) = "type") Or (LCase(sIdent) = "while") Or (LCase(sIdent) = "for") Or _
        (LCase(sIdent) = "loop") Or (LCase(sIdent) = "goto") Or (LCase(sIdent) = "label") Or (LCase(sIdent) = "binary") Or (LCase(sIdent) = "library") Or _
        (LCase(sIdent) = "exit") Or (LCase(sIdent) = "menu") Or (LCase(sIdent) = "dialog") Or (LCase(sIdent) = "accelerator") Or (LCase(sIdent) = "icon") Or _
        (LCase(sIdent) = "start") Or (LCase(sIdent) = "stop") Or (LCase(sIdent) = "import")
    If Not IsResIdent Then
        IsResIdent = SymbolTypeExists(sIdent)
    End If
End Function

Sub AddSymbol(Name As String, Size As Long, Section As SectionType, sType As SymbolType, TypeName As String, Optional Offset As Long = 0, Optional Index As Long = 0, Optional Parameter As Boolean = False, Optional Pointer As Boolean = False, Optional UserDefined As String = "")
    Dim sName As String
    sName = Name
    If IsResIdent(sName) Then errResIdentUsedAsIdent sName
    If sCurrentProc <> "" And Left(sName, 1) <> "$" Then sName = sCurrentProc & "." & sName
    If SymbolExists(sName) Or AliasExists(sName) Or ConstantExists(sName) Then
        errIdentifierExists sName
        Exit Sub
    End If
    ReDim Preserve Symbols(UBound(Symbols) + 1) As SYMBOL_ITEM
    With Symbols(UBound(Symbols))
        .Name = sName
        .Size = Size
        .Section = Section
        .sType = sType
        .TypeName = TypeName
        .Index = Index
        .Parameter = Parameter
        .Pointer = Pointer
        .UserDefined = UserDefined
        .ArgCount = -1
        If sType <> stLocal Then
            Select Case Section
            Case stIData
                If Offset = 0 Then
                    .Offset = UBound(IDataSection)
                Else
                    .Offset = Offset - 1
                End If
            Case stUData: .Offset = UDataSection
            Case stCode: .Offset = UBound(CodeSection)
            Case stImport: .Offset = UBound(ImportSection)
            Case stResource
                If Offset = 0 Then
                    .Offset = UBound(ResourceSection)
                Else
                    .Offset = Offset
                End If
            End Select
        Else
            .Offset = Offset
        End If
    End With
End Sub

Function SymbolExists(sName As String) As Boolean
    Dim sTemp As String
    If sCurrentProc <> "" Then sTemp = sCurrentProc & "." & sName
    SymbolExists = SymExists(sTemp)
    If Not SymbolExists Then SymbolExists = SymExists(sName)
End Function

Function SymExists(sName As String) As Boolean
    Dim i As Long
    SymExists = False
    For i = 1 To UBound(Symbols)
        With Symbols(i)
            If .Name = sName Then
                SymExists = True
                Exit Function
            End If
        End With
    Next i
End Function

Function GetSymbol(sName As String) As SYMBOL_ITEM
    Dim i As Long
    If SymExists(sCurrentProc & "." & sName) Then sName = sCurrentProc & "." & sName
    For i = 1 To UBound(Symbols)
        With Symbols(i)
            If .Name = sName Then
                GetSymbol = Symbols(i)
                Exit Function
            End If
        End With
    Next i
    errIdentNotFound sName
End Function

Sub SetArgCount(sName As String, lArgCount As Integer)
    Dim i As Long
    For i = 1 To UBound(Symbols)
        If Symbols(i).Name = sName Then Symbols(i).ArgCount = lArgCount
    Next i
End Sub
