Attribute VB_Name = "modSyntax"
Option Explicit

Public sUsedLibs As String, sTitle As String, sProductName As String, sDescription As String, sCopyright As String, sVersion As String, bAppLib As Boolean

Sub Application()
    Dim sIdent As String, bXP As Boolean
    sIdent = Identifier
    If Not LCase(sIdent) = "application" Then errUnexpectedEx "'application'", "'" & sIdent & "'"
    Blank
    sTitle = Identifier
    AddSymbol "App.Name", Len(sTitle), stIData, stGlobal, "char"
    AddIDataString sTitle
    Symbol (",")
    If IsNumber Then
        AddConstant "App.MajorVersion", Number
        Symbol "."
        AddConstant "App.MinorVersion", Number
        Symbol "."
        AddConstant "App.Revision", Number
        Symbol "."
        AddConstant "App.Build", Number
        sVersion = GetConstant("App.MajorVersion") & "." & GetConstant("App.MinorVersion") & "." & GetConstant("App.Revision") & "." & GetConstant("App.Build")
        Symbol (",")
        Select Case LCase(Identifier)
        Case "gui"
            AppType = atGUI
            AddConstant "App.Type", 1
        Case "console"
            AppType = atConsole
            AddConstant "App.Type", 0
        Case "cui"
            AppType = atConsole
            AddConstant "App.Type", 0
        Case Else
            errUnexpectedEx "'GUI', 'CUI' or 'Console'", "identifier"
        End Select
        bXP = False
        If IsSymbol(":") Then
            Symbol ":"
            If Not Identifier = "XP" Then errUnexpectedEx "'XP'", "identifier"
            AddAlias "InitCommonControls", "InitCommonControls", "COMCTL32.DLL", 0, False
            Invoke "InitCommonControls", 0
            bXP = True
        End If
        If IsSymbol("{") Then
            Symbol "{": SkipBlank
            If Not Identifier = "ProductName" Then errUnexpectedEx "'ProductName'", "identifier"
            Symbol ":"
            sProductName = Text
            AddSymbol "App.ProductName", Len(sProductName), stIData, stGlobal, "char"
            AddIDataString sProductName
            Terminator
            If Not Identifier = "Description" Then errUnexpectedEx "'Description'", "identifier"
            Symbol ":"
            sDescription = Text
            AddSymbol "App.Description", Len(sDescription), stIData, stGlobal, "char"
            AddIDataString sDescription
            Terminator
            If Not Identifier = "Copyright" Then errUnexpectedEx "'Copyright'", "identifier"
            Symbol ":"
            sCopyright = Text
            AddSymbol "App.Copyright", Len(sCopyright), stIData, stGlobal, "char"
            AddIDataString sCopyright
            Terminator
            Symbol "}": SkipBlank
            ResAddVersion
        Else
            AddSymbol "App.ProductName", 0, stIData, stGlobal, "char"
            AddSymbol "App.Description", 0, stIData, stGlobal, "char"
            AddSymbol "App.Copyright", 0, stIData, stGlobal, "char"
            AddIDataString ""
            Terminator
        End If
        bAppLib = False
        If bXP = True Then ResAddManifest
    Else
        If LCase(Identifier) <> "library" Then errUnexpected "'library'"
        bAppLib = True
        Terminator
    End If
End Sub

Sub Begin()
    Dim sIdent As String
    sIdent = Identifier
    If Not LCase(sIdent) = "begin" Then errUnexpectedEx "'begin'", "'" & sIdent & "'"
    SkipBlank
End Sub

Sub EndProc()
    Dim sIdent As String
    sIdent = Identifier
    If Not LCase(sIdent) = "end" Then errUnexpectedEx "'end'", "'" & sIdent & "'"
    Terminator
    SkipBlank
End Sub

Sub EndApp()
    Dim sIdent As String
    sIdent = Identifier
    If Not LCase(sIdent) = "end." Then errUnexpectedEx "'end.'", "'" & sIdent & "'"
    SkipBlank
    If Len(sCode) > 0 Then CompileError "code after 'end.' is ignored"
    Resources
End Sub

Sub GlobalData()
    Dim sIdent As String
    While Len(sCode) > 0
        If IsSymbol("?") Then Exit Sub
        If LCase(Locked(5)) = "begin" Then Exit Sub
        sIdent = Identifier
        If IsBlank Or IsSymbol("[") Then
            Select Case LCase(sIdent)
            Case "import": ImportData
            Case "include": IncludeFile
            Case "library": Library
            Case "alias": DeclareAlias
            Case "icon": ResAddIcon
            Case "bitmap": ResAddBitmap
            Case "menu": ResAddMenu
            Case "dialog": ResAddDialog
            Case "accelerator": ResAddAcc
            Case "manifest": ResAddManifest
            Case "version": ResAddVersion
            Case "const": DeclareConstant
            Case "frame": DeclareFrame
            Case "type": DeclareType
            Case "enum": DeclareEnum
            Case "binary": Binary
            Case "start"
                Blank
                ReDim Preserve LibBegin(UBound(LibBegin) + 1) As String
                LibBegin(UBound(LibBegin)) = Identifier
                Terminator
            Case "stop"
                Blank
                ReDim Preserve LibEnd(UBound(LibEnd) + 1) As String
                LibEnd(UBound(LibEnd)) = Identifier
                Terminator
            Case Else: DeclareData sIdent
            End Select
        Else
            sCode = sIdent & sCode
            Exit Sub
        End If
    Wend
End Sub

Sub LocalData()
    Dim sIdent As String
    While Len(sCode) > 0
        If LCase(Locked(5)) = "begin" Then Exit Sub
        sIdent = Identifier
        If IsBlank Or IsSymbol("[") Then
            DeclareData sIdent
        Else
            sCode = sIdent & sCode
            Exit Sub
        End If
    Wend
End Sub

Sub CodeBlock()
    Dim sIdent As String
    While Len(sCode) > 0
        If IsSymbol("}") Then Exit Sub
        If LCase(Locked(4)) = "end;" Or LCase(Locked(4)) = "end." Then Exit Sub
        sIdent = Identifier
        If LCase(sIdent) = "if" Then
            IfStatement
        ElseIf LCase(sIdent) = "while" Then
            WhileStatement
        ElseIf LCase(sIdent) = "for" Then
            ForStatement
        ElseIf LCase(sIdent) = "loop" Then
            LoopStatement
        ElseIf LCase(sIdent) = "return" Then
            ReturnProc
        ElseIf LCase(sIdent) = "goto" Then
            JumpToLabel
        ElseIf LCase(sIdent) = "label" Then
            DeclareLabel
        ElseIf LCase(sIdent) = "exit" Then
            Jump "$end"
            Terminator
        ElseIf LCase(sIdent) = "binary" Then
            Binary
        ElseIf IsAssignment Then
            sCode = sIdent & sCode
            Assignment
        ElseIf IsSymbol("=") Then
            errUnexpectedEx "':='", "'='"
        Else
            sCode = sIdent & sCode: Exit Sub
        End If
    Wend
End Sub

Sub Include(sFile As String, bLib As Boolean)
    Dim sOldFile As String, lOldLine As Long, sTemp As String, sTmp As String
    If bLib Then
        sTmp = mId(sFile, InStr(sFile, "\Include\") + 9)
        sTmp = Replace(Left(sTmp, Len(sTmp) - 4), "\", ".")
    End If
    If InStr(LCase(sUsedLibs), ";" & LCase(sTmp) & ";") = 0 Or Not bLib Then
        If bLib Then sUsedLibs = sUsedLibs & ";" & sTmp & ";"
        sCode = FileRead(sFile) & "?" & sCode
        sOldFile = sCurrentFile
        lOldLine = lCurrentLine
        sCurrentFile = sFile
        lCurrentLine = 1
        SkipBlank
        GlobalData
        Symbol "?"
        sCurrentFile = sOldFile
        lCurrentLine = lOldLine
    End If
End Sub

Sub IncDir(sDir As String)
    Dim i As Long, ii As Long, sTemp As String
    With frmMain
        sTemp = .dirs.Path
        .dirs.Path = sDir
        .Files.Path = sDir
        For i = 0 To .dirs.ListCount - 1
            IncDir .dirs.List(i)
        Next i
        For i = 0 To .Files.ListCount - 1
            Include sDir & "\" & .Files.List(i), True
        Next i
        .dirs.Path = sTemp
        .Files.Path = sTemp
    End With
End Sub

Sub IncludeFile()
    Blank
    Include Text, False
    Terminator
End Sub

Sub Library()
    Dim sLib As String, sTemp As String
    Blank
next_lib:
    sTemp = Identifier
    If InStr(LCase(sUsedLibs), ";" & LCase(sTemp) & ";") > 0 Then
        Terminator
        Exit Sub
    End If
    sLib = App.Path & "\..\Include\" & Replace(sTemp, ".", "\")
    If Not FileExists(sLib & ".lib") And Not DirExists(sLib) Then
        errUnknownLibrary sTemp
        Exit Sub
    End If
    If Not FileExists(sLib & ".lib") Then
        IncDir sLib
    Else
        Include sLib & ".lib", True
    End If
    If IsSymbol(",") Then
        Symbol ","
        GoTo next_lib
    End If
    Terminator
End Sub

Sub DeclareAlias()
    Dim sIdent As String, sFunc As String, sLib As String, lParams As Long, bPop As Boolean, sTemp As String
    Blank
    sIdent = Identifier
    Blank
    sTemp = Identifier
    If LCase(sTemp) = "in" Then
        sFunc = sIdent
    ElseIf LCase(sTemp) = "for" Then
        Blank
        sFunc = Text
        Blank
        sTemp = Identifier
    ElseIf LCase(sTemp) = "ascii" Then
        sFunc = sIdent & "A"
        Blank
        sTemp = Identifier
    ElseIf LCase(sTemp) = "unicode" Then
        sFunc = sIdent & "W"
        Blank
        sTemp = Identifier
    Else
        errUnexpected "'for' or 'ascii' or 'unicode'"
        Exit Sub
    End If
    If Not LCase(sTemp) = "in" Then
        errUnexpected "'in'"
        Exit Sub
    End If
    Blank
    sLib = Text
    lParams = -1
    bPop = False
    If Not IsTerminator Then
        Blank
        sTemp = Identifier
        If LCase(sTemp) = "paramcount" Then
            Symbol "="
            lParams = Number
            If Not IsTerminator Then
                Blank
                sTemp = Identifier
            End If
        End If
        If LCase(sTemp) = "restorestack" Then
            bPop = True
        End If
    End If
    If Not AliasExists(sIdent) Then
        AddAlias sIdent, sFunc, sLib, lParams, bPop
    End If
    Terminator
End Sub

Sub DeclareConstant()
    Dim sIdent As String
    Blank
    sIdent = Identifier
    Symbol "="
    AddConstant sIdent, Number
    Terminator
End Sub

Sub DeclareFrame()
    Dim ArgCount As Integer, sName As String, i As Long, sFuncName As String, sAlias As String
    Dim sTemp As String, lTemp As Long, st As SYMBOL_TYPE
    Blank
    sFuncName = Identifier
    lProcDataSize = 0: ArgCount = 0
    AddSymbol sFuncName, 0, stCode, stFunction, ""
    If IsBlank Then Blank
    sAlias = ""
    If IsIdentifier Then
        sAlias = LCase(Identifier)
        If sAlias = "alias" Then
            Blank
            sAlias = Identifier
            AddSymbol sAlias, 0, stCode, stFunction, ""
        Else
            errUnexpectedEx "'alias'", "'" & sAlias & "'"
        End If
    End If
    sCurrentProc = sFuncName
    Symbol "("
    While Not IsSymbol(")")
        If Len(sCode) = 0 Then Exit Sub
        sTemp = Identifier
        Blank
        If GetTypeSize(sTemp) = 0 And sTemp <> "char" Then
            If Not SymbolTypeExists(sTemp) Then
                errInvalidType sTemp
                Exit Sub
            End If
            sName = Identifier
            AddSymbol sName, 4, stUData, stLocal, "char", 8 + (ArgCount * 4), 0, True, , sTemp
            AddConstant sCurrentProc & "." & sName & ".Size", st.Size
            st = GetSymbolType(sTemp)
            lTemp = 0
            For i = LBound(st.SubTypes) + 1 To UBound(st.SubTypes)
                With st.SubTypes(i)
                If GetTypeSize(.TypeName) > 0 Or .TypeName = "char" Then
                    AddSymbol sName & "." & .Name, .Size, stUData, stLocal, .TypeName, 8 + (ArgCount * 4), lTemp, True, True
                    lTemp = lTemp + .Size
                Else
                    errInvalidType .TypeName
                End If
                End With
            Next i
            ArgCount = ArgCount + 1
        Else
            If sTemp = "word" Or sTemp = "byte" Then sTemp = "dword"
            AddSymbol Identifier, 4, stUData, stLocal, sTemp, 8 + (ArgCount * 4), 0, True
            ArgCount = ArgCount + 1
        End If
        If Not IsSymbol(")") Then Symbol (",")
    Wend
    SetArgCount sFuncName, ArgCount
    If sAlias <> "" Then SetArgCount sAlias, ArgCount
    Symbol ")"
    Terminator
    LocalData
    Begin
    StartFrame
    LocalDataSize lProcDataSize
    CodeBlock
    EndProc
    AddSymbol "endproc", 0, stCode, stFunction, ""
    EndFrame ArgCount * 4
    sCurrentProc = ""
End Sub

Sub ReturnProc()
    If sCurrentProc = "" Then CompileError "return is not inside a frame"
    Symbol "("
    Expression
    Symbol ")"
    Jump sCurrentProc & ".endproc"
    Terminator
End Sub

Sub DeclareType()
    Dim sName As String, lSize As Long, sTypes As String
    Dim sTemp As String, lTemp As Long, sTempName As String
    Blank
    sName = Identifier
    lSize = 0: sTypes = ""
    Symbol "{": SkipBlank
    If IsSymbol("}") Then errEmptyType
    While Not IsSymbol("}")
        sTemp = Identifier
        sTypes = sTypes & sTemp
        If IsSymbol("[") And sTemp = "char" Then
            Symbol "["
            lTemp = Number
            Symbol "]", False
            sTypes = sTypes & ":" & lTemp & ";"
            lSize = lSize + lTemp
        Else
            sTypes = sTypes & ";"
            lSize = lSize + GetTypeSize(sTemp)
        End If
        Blank
        sTempName = Identifier
        If sTempName = "Size" Then errUnexpectedEx "identifier", "'" & sTempName & "'"
        sTypes = sTypes & sTempName & ";"
        If IsSymbol("=") Then
            Symbol "="
            If IsIdentifier And sTemp = "ptr" Then
                sTypes = sTypes & Identifier & ";"
            ElseIf IsNumber And sTemp <> "ptr" And sTemp <> "char" Then
                sTypes = sTypes & Number & ";"
            End If
        Else
            sTypes = sTypes & ";"
        End If
        Terminator
        If Len(sCode) = 0 Then Exit Sub
    Wend
    AddSymbolType sName, lSize, sTypes
    AddConstant sName & ".Size", lSize
    Symbol "}": SkipBlank
End Sub

Sub DeclareEnum()
    Symbol "{": SkipBlank
    AddEnum Identifier
    While IsSymbol(",")
        Symbol (",")
        SkipBlank
        AddEnum Identifier
    Wend
    SkipBlank
    Symbol "}": SkipBlank
End Sub

Sub IfStatement()
    Dim lId As Long, sTemp As String, lCount As Long
    lId = lUniqueID
    lUniqueID = lUniqueID + 1
    lCount = 0
    Symbol "("
    Expression
    Symbol ")"
    Expr_Test "$elseif" & lId & "-" & lCount
    Symbol "{": SkipBlank
    CodeBlock
    Symbol "}": SkipBlank
    Jump "$endif" & lId
newblock:
    If Not IsIdentifier Then GoTo endblock
    sTemp = Identifier
    Select Case LCase(sTemp)
    Case "else"
        AddSymbol "$elseif" & lId & "-" & lCount, 0, stCode, stFunction, ""
        lCount = lCount + 1
        Symbol "{": SkipBlank
        CodeBlock
        Symbol "}": SkipBlank
        Jump "$endif" & lId
    Case "elseif"
        AddSymbol "$elseif" & lId & "-" & lCount, 0, stCode, stFunction, ""
        lCount = lCount + 1
        Symbol "("
        Expression
        Expr_Test "$elseif" & lId & "-" & lCount
        Symbol ")"
        Symbol "{": SkipBlank
        CodeBlock
        Symbol "}": SkipBlank
        Jump "$endif" & lId
    Case Else
        sCode = sTemp & sCode
endblock:
        AddSymbol "$elseif" & lId & "-" & lCount, 0, stCode, stFunction, ""
        AddSymbol "$endif" & lId, 0, stCode, stFunction, ""
        Exit Sub
    End Select
    GoTo newblock
End Sub

Sub WhileStatement()
    Dim lId As Long
    lId = lUniqueID
    lUniqueID = lUniqueID + 1
    AddSymbol "$while" & lId, 0, stCode, stFunction, ""
    Symbol "("
    Expression
    Symbol ")"
    Expr_Test "$endwhile" & lId
    Symbol "{": SkipBlank
    CodeBlock
    Jump "$while" & lId
    AddSymbol "$endwhile" & lId, 0, stCode, stFunction, ""
    Symbol "}": SkipBlank
End Sub

Sub ForStatement()
    Dim lId As Long, sIdent As String, bInc As Boolean
    lId = lUniqueID
    lUniqueID = lUniqueID + 1
    Symbol "("
    sIdent = Identifier
    sCurrType = GetSymbol(sIdent).TypeName
    GetAddress sIdent
    Expr_Push
    Symbol ":=": Expression: Store
    Symbol ";"
    AddSymbol "$for" & lId, 0, stCode, stFunction, ""
    Expression
    Expr_Test "$endfor" & lId
    Symbol ";"
    sIdent = Identifier
    sCurrType = GetSymbol(sIdent).TypeName
    GetAddress sIdent
    Expr_Push
    If IsSymbol("++") Then
        bInc = True
        Symbol "++"
    Else
        bInc = False
        Symbol "--"
    End If
    Symbol ")"
    Symbol "{": SkipBlank
    CodeBlock
    If bInc Then
        Increase
    Else
        Decrease
    End If
    Jump "$for" & lId
    AddSymbol "$endfor" & lId, 0, stCode, stFunction, ""
    Symbol "}": SkipBlank
End Sub

Sub LoopStatement()
    Dim lId As Long
    lId = lUniqueID
    lUniqueID = lUniqueID + 1
    Symbol "("
    Expression
    Expr_MovECXEAX
    Symbol ")"
    AddSymbol "$loop" & lId, 0, stCode, stFunction, ""
    Symbol "{": SkipBlank
    Expr_PushECX
    CodeBlock
    Expr_PopECX
    Expr_DecECX
    Expr_TestECX "$loop" & lId
    Symbol "}": SkipBlank
End Sub

Sub JumpToLabel()
    Blank
    If Len(sCurrentProc) = 0 Then
        Jump Identifier
    Else
        Jump sCurrentProc & "." & Identifier
    End If
    Terminator
End Sub

Sub DeclareLabel()
    Blank
    AddSymbol Identifier, 0, stCode, stFunction, ""
    Terminator
End Sub

Sub Binary()
    Blank
    AddSymbol Identifier, 0, stCode, stFunction, ""
    Symbol (",")
    FileBinary Text
    Terminator
End Sub

Sub ImportData()
    On Error GoTo err
    Dim sFile As String, FileNum As Long
    Dim NumItems As Long, i As Long
    Blank
    sFile = Text
    If mId(sFile, 2, 1) <> ":" And Trim(sCurrentFile) <> "" Then
        sFile = GetFilePath(GetFullFilename(sCurrentFile)) & "\" & sFile
    End If
    If Not FileExists(sFile) Then
        errFileNotFound GetFileTitle(sFile)
        Exit Sub
    Else
        FileNum = FreeFile
        Open GetFullFilename(sFile) For Binary As #FileNum
    End If
    ' Alias
    Get #FileNum, , NumItems
    ReDim Preserve Alias(UBound(Alias) + NumItems) As ALIAS_ITEM
    For i = 1 To NumItems
        Get #FileNum, , Alias(UBound(Alias) - NumItems + i)
    Next i
    ' Constant
    Get #FileNum, , NumItems
    ReDim Preserve Constants(UBound(Constants) + NumItems) As CONSTANT_ITEM
    For i = 1 To NumItems
        Get #FileNum, , Constants(UBound(Constants) - NumItems + i)
    Next i
    ' Type
    Get #FileNum, , NumItems
    ReDim Preserve SymbolTypes(UBound(SymbolTypes) + NumItems) As SYMBOL_TYPE
    For i = 1 To NumItems
        Get #FileNum, , SymbolTypes(UBound(SymbolTypes) - NumItems + i)
    Next i
    Close #FileNum
    Terminator
    Exit Sub
err:
    errInvalidFileFormat
End Sub

Sub ExportData(sFile As String)
    Dim FileNum As Long, i As Long
    If mId(sFile, 2, 1) <> ":" And Trim(sCurrentFile) <> "" Then
        sFile = GetFilePath(GetFullFilename(sCurrentFile)) & sFile
    End If
    KillIfExists sFile
    FileNum = FreeFile
    Open GetFullFilename(sFile) For Binary As #FileNum
    ' Alias
    Put #FileNum, , UBound(Alias)
    For i = 1 To UBound(Alias)
        Put #FileNum, , Alias(i)
    Next i
    ' Constant
    Put #FileNum, , UBound(Constants)
    For i = 1 To UBound(Constants)
        Put #FileNum, , Constants(i)
    Next i
    ' Type
    Put #FileNum, , UBound(SymbolTypes)
    For i = 1 To UBound(SymbolTypes)
        Put #FileNum, , SymbolTypes(i)
    Next i
    Close #FileNum
End Sub
