Attribute VB_Name = "modTypes"
Option Explicit

Public Type SUBTYPE_ITEM
    Name As String
    TypeName As String
    Size As Long
    lValue As Long
    sValue As String
End Type

Public Type SYMBOL_TYPE
    Name As String
    Size As Long
    SubTypes() As SUBTYPE_ITEM
End Type

Public SymbolTypes() As SYMBOL_TYPE

Sub InitSymbolTypes()
    ReDim Preserve SymbolTypes(0) As SYMBOL_TYPE
    ReDim Preserve SymbolTypes(0).SubTypes(0) As SUBTYPE_ITEM
End Sub

Sub AddSymbolType(sName As String, lSize As Long, sSubtypes As String)
    Dim i As Long, sTemp() As String, lTempSize As Long
    If SymbolTypeExists(sName) Then errTypeExists sName
    ReDim Preserve SymbolTypes(UBound(SymbolTypes) + 1) As SYMBOL_TYPE
    ReDim Preserve SymbolTypes(UBound(SymbolTypes)).SubTypes(0) As SUBTYPE_ITEM
    With SymbolTypes(UBound(SymbolTypes))
        .Name = sName
        .Size = lSize
        lTempSize = 0
        sTemp = Split(sSubtypes, ";")
        For i = LBound(sTemp) To UBound(sTemp) - 1 Step 3
            ReDim Preserve .SubTypes(UBound(.SubTypes) + 1) As SUBTYPE_ITEM
            With .SubTypes(UBound(.SubTypes))
                .Name = sTemp(i + 1)
                AddConstant sName & "." & .Name, lTempSize
                .TypeName = sTemp(i)
                .Size = GetTypeSize(sTemp(i))
                If .Size = 0 And Left(.TypeName, 5) = "char:" Then
                    .Size = mId(.TypeName, InStr(.TypeName, ":") + 1)
                    .TypeName = "char"
                End If
                If sTemp(i + 2) <> "" Then
                    If InStr("-0123456789", Left(sTemp(i + 2), 1)) > 0 Then
                        .lValue = CLng(sTemp(i + 2))
                        .sValue = ""
                    Else
                        .sValue = sTemp(i + 2)
                        .lValue = 0
                    End If
                End If
                lTempSize = lTempSize + .Size
            End With
        Next i
    End With
End Sub

Function SymbolTypeExists(sName As String) As Boolean
    Dim i As Long
    SymbolTypeExists = (GetTypeSize(sName) > 0) Or (sName = "char")
    If SymbolTypeExists Then Exit Function
    For i = 1 To UBound(SymbolTypes)
        If SymbolTypes(i).Name = sName Then
            SymbolTypeExists = True
            Exit Function
        End If
    Next i
End Function

Function GetSymbolType(sName As String) As SYMBOL_TYPE
    Dim i As Long
    If Not SymbolTypeExists(sName) Then errTypeNotFound sName
    For i = 1 To UBound(SymbolTypes)
        If SymbolTypes(i).Name = sName Then
            GetSymbolType = SymbolTypes(i)
            Exit Function
        End If
    Next i
End Function

Function GetTypeSize(sType As String) As Long
    Select Case sType
    Case "byte": GetTypeSize = 1
    Case "word": GetTypeSize = 2
    Case "dword": GetTypeSize = 4
    Case "int": GetTypeSize = 4
    Case "ptr": GetTypeSize = 4
    Case Else: GetTypeSize = 0
    End Select
End Function

Sub DeclareData(sType As String)
    If Not SymbolTypeExists(sType) Then errUnexpectedEx "type identifier", "'" & sType & "'"
    If GetTypeSize(sType) > 0 Then
        PredefinedType GetTypeSize(sType), sType
    ElseIf sType = "char" Then
        Buffer
    Else
        UserdefinedType sType
    End If
End Sub

Sub AddLocalData(sName As String, lSize As Long, sType As String)
    lProcDataSize = lProcDataSize + lSize
    AddSymbol sName, lSize, stUData, stLocal, sType, -lProcDataSize
End Sub

Sub AddGlobalUData(sName As String, lSize As Long, sType As String)
    AddSymbol sName, lSize, stUData, stGlobal, sType
    AddUDataBuffer lSize
End Sub

Sub AddGLUData(sName As String, lSize As Long, sType As String)
    If Len(sCurrentProc) > 0 Then
        AddLocalData sName, lSize, sType
    Else
        AddGlobalUData sName, lSize, sType
    End If
End Sub

Sub AddGlobalIData(sName As String, lSize As Long, sType As String, lValue As Long)
    Dim i As Long
    AddSymbol sName, lSize, stIData, stGlobal, sType
    Select Case lSize
    Case 1
        If Not IsByte(lValue) Then errValueOutOfRange
        AddIDataByte CByte(lValue)
    Case 2
        If Not IsWord(lValue) Then errValueOutOfRange
        AddIDataWord CInt(lValue)
    Case 4
        AddIDataDWord lValue
    Case Else
        For i = 1 To lSize
            AddIDataByte 0
        Next i
    End Select
End Sub

Sub PredefinedType(lSize As Long, sType As String)
    Dim sTemp As String, sValue As String
    Blank
    sTemp = Identifier
    If IsSymbol("=") And Len(sCurrentProc) = 0 Then
        Symbol "="
        If sType = "ptr" Then
            AddSymbol sTemp, lSize, stIData, stGlobal, sType
            sValue = Identifier
            If AliasExists(sValue) Then
                errAliasAsPtrValue
            Else
                AddFixup sValue, stIData, UBound(IDataSection) + 1, 0
            End If
            AddIDataDWord 0
        Else
            AddGlobalIData sTemp, lSize, sType, Number
        End If
    Else
        AddGLUData sTemp, lSize, sType
    End If
    Terminator
End Sub

Sub Buffer()
    Dim sName As String, sValue As String, lValue As Long
    If IsSymbol("[") Or Len(sCurrentProc) > 0 Then
        Symbol "["
        lValue = Number
        Symbol "]", False
        Blank
        sName = Identifier
        If Len(sCurrentProc) > 0 Then
            lProcDataSize = Align4(lProcDataSize + lValue)
            AddSymbol sName, lValue, stUData, stLocal, "char", -lProcDataSize
        Else
            AddSymbol sName, lValue, stUData, stGlobal, "char"
            AddUDataBuffer lValue
        End If
    Else
        Blank
        sName = Identifier
        Symbol "="
        If IsSymbol("{") Then
            Symbol "{"
            AddSymbol sName, 1, stIData, stGlobal, "char"
            If IsText Then
                AddIDataString Text
            Else
                lValue = Number
                If Not IsByte(lValue) Then errValueOutOfRange
                AddIDataByte CByte(lValue)
            End If
            While Not IsSymbol("}")
                Symbol (",")
                If IsText Then
                    AddIDataString Text
                Else
                    lValue = Number
                    If Not IsByte(lValue) Then errValueOutOfRange
                    AddIDataByte CByte(lValue)
                End If
                If Len(sCode) = 0 Then Exit Sub
            Wend
            Symbol "}"
        Else
            sValue = Text
            AddSymbol sName, Len(sValue), stIData, stGlobal, "char"
            AddIDataString sValue
        End If
    End If
    Terminator
End Sub

Sub UserdefinedType(sType As String)
    Dim sName As String, st As SYMBOL_TYPE, i As Long, bInit As Boolean
    Dim sTemp As String, ii As Long, lOffset As Long
    st = GetSymbolType(sType)
    If bError Then Exit Sub
    Blank
    sName = Identifier
    bInit = IsSymbol("=")
    If bInit And Len(sCurrentProc) = 0 Then
        Symbol "="
        Symbol "{"
        AddSymbol sName, st.Size, stIData, stGlobal, "char", , , , , sType
    Else
        bInit = False
        Terminator
        If Len(sCurrentProc) > 0 Then
            lOffset = Align4(st.Size)
            AddSymbol sName, st.Size, stUData, stLocal, "char", -lProcDataSize - lOffset, , , , sType
        Else
            AddSymbol sName, st.Size, stUData, stGlobal, "char", , , , , sType
        End If
    End If
    For i = LBound(st.SubTypes) + 1 To UBound(st.SubTypes)
        With st.SubTypes(i)
        If GetTypeSize(.TypeName) > 0 Or .TypeName = "char" Then
            If .lValue <> 0 Or .sValue <> "" And bInit Then
                If .TypeName = "ptr" Then
                    AddSymbol sName & "." & .Name, .Size, stIData, stGlobal, .TypeName
                    AddFixup .sValue, stIData, UBound(IDataSection) + 1, 0
                    AddIDataDWord 0
                Else
                    AddGlobalIData sName & "." & .Name, .Size, .TypeName, .lValue
                End If
            ElseIf bInit And (IsIdentifier Or IsNumber Or IsText) Then
                If .TypeName = "char" Then
                    AddSymbol sName & "." & .Name, .Size, stIData, stGlobal, .TypeName
                    sTemp = Text
                    If Len(sTemp) >= .Size Then errValueOutOfRange
                    AddIDataString sTemp
                    For ii = 1 To .Size - Len(sTemp) - 1
                        AddIDataByte 0
                    Next ii
                ElseIf .TypeName = "ptr" Then
                    AddSymbol sName & "." & .Name, .Size, stIData, stGlobal, .TypeName
                    AddFixup Identifier, stIData, UBound(IDataSection) + 1, 0
                    AddIDataDWord 0
                Else
                    AddGlobalIData sName & "." & .Name, .Size, .TypeName, Number
                End If
                If Not IsSymbol("}") Then Symbol (",")
            ElseIf bInit Then
                AddGlobalIData sName & "." & .Name, .Size, .TypeName, 0
            Else
                If Len(sCurrentProc) = 0 Then
                    AddGlobalUData sName & "." & .Name, .Size, .TypeName
                Else
                    lOffset = lOffset - .Size
                    AddSymbol sName & "." & .Name, .Size, stUData, stLocal, .TypeName, -lProcDataSize - lOffset - .Size
                End If
            End If
        Else
            errInvalidType .TypeName
        End If
        End With
    Next i
    If bInit Then
        Symbol "}"
        Terminator
    End If
    If Len(sCurrentProc) > 0 Then lProcDataSize = lProcDataSize + st.Size
End Sub

Function Align4(lValue As Long) As Long
    If Int(lValue / 4) * 4 = lValue Then
        Align4 = lValue
    Else
        Align4 = (Int(lValue / 4) + IIf(lValue < 0, 0, 1)) * 4
    End If
End Function
